import numpy as np
import logging
import zipfile
import os
import shutil
from sentence_transformers import SentenceTransformer, LoggingHandler
from sys import exit

class TrainSentences:

    def __init__(self, txt_file):
        dir_path = os.getcwd() + '/'
        file_path = dir_path + txt_file
        print('\n')
        print('Scanning the path '+file_path+ ' ...')
        self.pretrain_model = 'bert-base-nli-mean-tokens'
        print('Pretrained model is set to '+self.pretrain_model+ ' ...')
        if(os.path.isfile(file_path) and self.get_file_extension(file_path) == ".txt"):
            print('Training file validation OK...')
            self.train_file_path = file_path
            if not os.path.exists(dir_path+'trained_model'):
                os.makedirs(dir_path+'trained_model')
            self.model_save_path = dir_path+'trained_model/'
            self.zip_save_path = dir_path+'/'
        else:
            exit('Training file is not valid... exiting...')
    
    def pretrained_model(self,model_name):
        self.pretrain_model = model_name
        print('Pretrained model is reset to '+model_name+ ' ...')

    def get_file_extension(self,src):
        return os.path.splitext(src)[-1].lower()

    def get_path(self):
        _vector_file = 'vector.npy'
        _train_file = 'train.txt'
        _files = {
            'model': self.model_save_path,
            'vector': self.model_save_path + _vector_file,
            'training_set': self.train_file_path,
            'zip_path' : self.zip_save_path+'model.zip',
            'train_file' : self.model_save_path + _train_file,
        }
        return _files

    def train(self):
        path = self.get_path()
        np.set_printoptions(threshold=100)
        logging.basicConfig(format='%(asctime)s - %(message)s',
                            datefmt='%Y-%m-%d %H:%M:%S',
                            level=logging.ERROR,
                            handlers=[LoggingHandler()])
        model = SentenceTransformer(self.pretrain_model)
        sentences = open(path.get('training_set')).read().splitlines()
        sentence_embeddings = model.encode(sentences)
        vecs = np.stack(sentence_embeddings)
        model.save(path.get('model'))
        print('Saving the model to '+path.get('model')+'...')
        np.save(path.get('vector'), sentence_embeddings)
        print('Saving the vector to '+path.get('vector')+'...')
        print('Initiating model compression(.zip) ...')
        os.rename(path.get('training_set'), path.get('train_file'))
        self.compress_file(path.get('model'),path.get('zip_path'))
        print('→ Download "model.zip" and use it for prediction ...')
       
    def compress_file(self,dirpath, zippath):
        fzip = zipfile.ZipFile(zippath, 'w', zipfile.ZIP_DEFLATED)
        basedir = os.path.dirname(dirpath) + '/' 
        for root, dirs, files in os.walk(dirpath):
            dirname = root.replace(basedir, '')
            for f in files:
                fzip.write(root + '/' + f, dirname + '/' + f)
        fzip.close()
        shutil.rmtree(self.model_save_path)


import numpy as np
import scipy.spatial
import logging
import json
import nltk
import zipfile
import os
import xlsxwriter
from sentence_transformers import SentenceTransformer, LoggingHandler
from .TrainSentences import TrainSentences
from sys import exit


class SimilarSentences:

    def __init__(self, path, type):
        logging.basicConfig(level=logging.ERROR)
        nltk.download('punkt')
        if(type == "predict"):
            self.path = path
            self.load()
        elif(type == "train"):
            self.train_file = path

    def load(self):
        dir_path = os.getcwd()
        model_path = dir_path+'/model/'
        self.model_path = model_path
        if(not os.path.isdir(model_path+'0_BERT') and not os.path.isdir(model_path+'1_Pooling') and not os.path.isfile(model_path+'vector.npy')):
            self.reload()
        else:
            print('For reloading/updating the model try model.relaod()')

    def reload(self):
        dir_path = os.getcwd()
        model_path = dir_path+'/model/'
        print('Scanning the path '+model_path + ' ...')
        path = self.path
        if(zipfile.is_zipfile(path)):
            with zipfile.ZipFile(path, 'r') as zip_ref:
                zip_ref.extractall(model_path)
                if(os.path.isdir(model_path+'0_BERT') and os.path.isdir(model_path+'1_Pooling') and os.path.isfile(model_path+'vector.npy')):
                    print('Model validation OK...')
                else:
                    exit('Model file is not valid... exiting...')
        else:
            exit('Model file is not in .zip format')

    def get_path(self):
        _vector_file = 'vector.npy'
        _training_set = 'train.txt'
        _files = {
            'model': self.model_path,
            'vector': self.model_path + _vector_file,
            'training_set': self.model_path+_training_set
        }
        return _files

    def get_sentences(self, text):
        return nltk.sent_tokenize(text)

    def predict(self, text, num_of_simlar_sentences, response="simple"):
        path = self.get_path()
        model = SentenceTransformer(path.get('model'))
        given_text = self.get_sentences(text)
        text_embedding = model.encode(given_text)
        return self.output(given_text, text_embedding, num_of_simlar_sentences, response)

    def output(self, text, embedding, closest_n, response):

        path = self.get_path()
        sentences = open(path.get('training_set')).read().splitlines()
        sentences_vector = np.load(path.get('vector'))
        vectors = np.stack(sentences_vector)
        output = {}
        i = 0
        while i < closest_n:
            out = []
            for query, sentence in zip(text, embedding):
                distances = scipy.spatial.distance.cdist(
                    [sentence], vectors, "cosine")[0]

                results = zip(range(len(distances)), distances)
                results = sorted(results, key=lambda x: x[1])
                index = 0
                for idx, distance in results[0:closest_n]:
                    if(i == index):
                        out.append(
                            {"sentence": sentences[idx].strip(), "score": (1-distance)})
                    index += 1
                output[i] = out
            i += 1
        return self.json_output(output, response)

    def json_output(self, data, response):
        data = [*data.values()]
        json_out = ''
        if(response == "detailed"):
            json_out = json.dumps(data)
        elif(response == "simple"):
            simple_data = []
            for i in data:
                str = ''
                k = 0
                for j in i:
                    if(k == 0):
                        str += j['sentence']
                    else:
                        str += ' '+j['sentence']
                    k += 1
                simple_data.append(str)
            json_out = json.dumps(simple_data)
        return json_out
    
    def load_pretrained_model(self,model_name):
        model = TrainSentences(self.train_file)
        model.pretrained_model(model_name)

    def train(self):
        model = TrainSentences(self.train_file)
        model.train()

    def batch_output(self, type: str = None):
        if(type == None):
            type = 'excel'
        
        if(type == 'excel'):
            output = self.create_excel()
    
    def create_excel(self):
        workbook = xlsxwriter.Workbook('Results.xlsx') 
        # Some data we want to write to the worksheet. 
        sentences_scores = ( 
            ['Given Sentence',"Suggestion(1)", "Score(1)"], 
        ) 
        row = 0
        col = 0
        
        # Iterate over the data and write it out row by row. 
        for input_sentence, output_sentence, score in (sentences_scores): 
            worksheet.write(row, col, input_sentence) # Given sentence
            worksheet.write(row, col + 1, output_sentence) # Suggested Sentence
            worksheet.write(row, col + 2, score) # Score
            row += 1
        
        workbook.close() 




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="similar-sentences",
    version="2.2",
    author="Shankar Ganesh Jayaraman",
    author_email="shangan.23@gmail.com",
    description="Similar sentence prediction",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/shangan23/similar-sentences",
    packages=setuptools.find_packages(),
    install_requires=[
        "sentence-transformers",
        "xlsxwriter"
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)
